/*
 * Icon configuration.
 */

/*
 * Define things that are needed in define.h.
 */

/*
 * Representations of directories. LocalDir is the "current working directory".
 *  SourceDir is where the source file is. 
 */

#define LocalDir ""
#define SourceDir (char *)NULL

/*
 * System-specific definitions are in define.h
 */
#include "../h/define.h"

#include <stdio.h>

/*
 *  A number of symbols are defined here.  Some are specific to individual
 *  to operating systems.  Examples are:
 *
 *	MSDOS		MS-DOS for PCs
 *	UNIX		any UNIX system
 *	VMS		VMS for the VAX
 *
 *  These are defined to be 1 or 0 depending on which operating system
 *  the installation is being done under.  They are all defined and only
 *  one is defined to be 1.  (They are used in the form #if VAX || MSDOS.)
 *
 *  There also are definitions of symbols for specific computers and
 *  versions of operating systems.  These include:
 *
 *	SUN		code specific to the Sun Workstation
 *	LATTICE		code specific to the Lattice C compiler for MS-DOS
 *
 *  Other definitions may occur for special configurations. These include:
 *
 *	DeBug		debugging code
 *	MemMon		memory monitoring output
 *	TranStats	translator statistics gathering
 *
 *  Other definitions perform configurations that are common to several
 *  systems. Examples are:
 *
 *	Double		align reals at double-word boundaries
 *	ZeroDivide	enable C code to test for division by real zero.
 *
 */

/*
 * The following definitions insure that all the symbols for operating
 * systems that are not relevant are defined to be 0 -- so that they
 * can be used in logical expressions in #if directives.
 */

#ifndef PORT
#define PORT 0
#endif					/* PORT */

#ifndef AMIGA
#define AMIGA 0
#endif					/* AMIGA */

#ifndef ATARI_ST
#define ATARI_ST 0
#endif					/* ATARI_ST */

#ifndef HIGHC_386
#define HIGHC_386 0
#endif					/* HIGHC_386 */

#ifndef MACINTOSH
#define MACINTOSH 0
#endif					/* MACINTOSH */

#ifndef MSDOS
#define MSDOS 0
#endif					/* MSDOS */

#ifndef MVS
#define MVS 0
#endif					/* MVS */

#ifndef OS2
#define OS2 0
#endif					/* OS2 */

#ifndef UNIX
#define UNIX 0
#endif					/* UNIX */

#ifndef VM
#define VM 0
#endif					/* VM */

#ifndef VMS
#define VMS 0
#endif					/* VMS */


/*
 * The following definitions serve to cast common conditionals is
 *  a positive way, while allowing defaults for the cases that
 *  occur most frequently.  That is, if co-expressions are not supported,
 *  NoCoexpr is defined in define.h, but if they are supported, no
 *  definition is needed in define.h; nonetheless subsequent conditionals
 *  can be cast as #ifdef Coexpr.
 */

#ifndef NoCoexpr
#define Coexpr
#endif					/* NoCoexpr */

#ifndef NoOver
#define OverCheck
#endif					/* NoOver */

#ifndef NoEnvVars
#define EnvVars
#endif					/* NoEnvVars */

#ifndef NoTraceBack
#define TraceBack
#endif					/* NoTraceBack */

#ifndef NoStrInvoke
#define StrInvoke
#endif					/* NoStrInvoke */

#ifndef NoHeader
#define Header
#endif					/* NoHeader */

/*
 * Other defaults.
 */

#ifndef AllocType
#define AllocType unsigned int
#endif					/* AllocType */

typedef AllocType msize;

#ifndef ErrorExit
#define ErrorExit 1
#endif					/* ErrorExit */

#ifndef NormalExit
#define NormalExit 0
#endif					/* NormalExit */

#ifndef Hz
#define Hz 60
#endif					/* Hz */

#ifndef MaxHdr
#define MaxHdr 4096
#endif					/* MaxHdr */

#ifndef SysTime
#define SysTime <time.h>
#endif					/* SysTime */

#ifndef WordBits
#define WordBits 32
#endif					/* WordBits */

#ifndef IntBits
#define IntBits WordBits
#endif					/* IntBits */

#ifndef SourceSuffix
#define SourceSuffix ".icn"
#endif					/* SourceSuffix */

#ifndef IcodeSuffix
#define IcodeSuffix ""
#endif					/* IcodeSuffix */

#ifndef IcodeASuffix
#define IcodeASuffix ""
#endif					/* IcodeASuffix */

#ifndef U1Suffix
#define U1Suffix ".u1"
#endif					/* U1Suffix */

#ifndef U2Suffix
#define U2Suffix ".u2"
#endif					/* U2Suffix */

#ifndef USuffix
#define USuffix ".u"
#endif					/* USuffix */

#ifndef TargetDir
#define TargetDir LocalDir
#endif					/* TargetDir */

#ifndef Options
#if UNIX
#define Options "ce:mo:stuxLS:"
#else					/* UNIX */
#define Options "ce:o:stuxLS:"
#endif					/* UNIX */
#endif					/* Options */

#ifndef Usage
#if UNIX
#define Usage "[-cmstu] [-e efile] [-o ofile] [-Sxnnnn]"
#else					/* UNIX */
#define Usage "[-cstu] [-e efile] [-o ofile] [-Sxnnnn]"
#endif					/* UNIX */
#endif					/* Usage */

#if AMIGA || ATARI_ST || MSDOS || MVS || UNIX || VM || VMS
#define SystemFncs
#endif					/* AMIGA || ATARI_ST || ... */

/*
 * Default sizing and such.
 */

/*
 * Set up typedefs and related definitions depending on whether or not
 * ints and pointers are the same size.
 */

#if IntBits == 16
typedef long int word;
typedef unsigned long int uword;
#define sbrk lsbrk
#define qsort llqsort
#define strlen lstrlen
#else					/* WordBits != IntBits */
typedef int word;
typedef unsigned int uword;
#endif					/* WordBits != IntBits */

/*
 * Provide definition to use void if it's supported (mainly avoids error
 *  messages with some C compilers.  Note: typedef does not work, so care
 *  is needed not to use compound declarations.
 */

#ifdef VoidType
#define novalue void
#define noargs void
#else					/* VoidType */
#define novalue int
#define noargs
#endif					/* VoidType */

#define WordSize sizeof(word)

#ifndef ByteBits
#define ByteBits 8
#endif					/* ByteBits */

/*
 * Change the name of gcvt() if we're supplying our own version,
 * to avoid complaints under VMS and others with shared libraries.
 */

#ifdef IconGcvt
#define gcvt icon_gcvt
#endif					/* IconGcvt */


/*
 *  The following definitions depend on whether or not the ANSI C standard
 *  is supported.
 */

#ifdef Standard
#define Cat(x,y) x##y
#define Lit(x) #x
#ifndef PointerDef
typedef void *pointer;
#endif					/* PointerDef */
#ifndef VoidType
#define VoidType
#endif					/* VoidType */
#else					/* Standard */
#define Ident(x) x
#define Cat(x,y) Ident(x)y
#define Lit(x) "x"
#ifndef PointerDef
typedef char *pointer;
#endif					/* PointerDef */
#endif					/* Standard */

/*
 * Customize output if not pre-defined.
 */

#ifndef TraceOut
#define TraceOut(s) fprintf(stderr,s)
#endif					/* TraceOut */

#ifdef EBCDIC
#define BackSlash "\xe0"
#else					/* EBCDIC */
#define BackSlash "\\"
#endif					/* EBCDIC */

/*
 * Typedefs to make some things easier.
 */

typedef int (*fptr)();
typedef struct descrip *dptr;

/*
 * Function prototypes.
 */

#include "../h/proto.h"

/*
 * The following code is operating-system dependent [@config.01].
 *  Any configuration stuff that has to be done at this point.
 */

#if PORT
   /* Probably nothing is needed. */
Deliberate Syntax Error
#endif					/* PORT */

#if AMIGA || ATARI_ST || MACINTOSH || MSDOS || MVS || UNIX || VM || VMS
   /* Nothing is needed */
#endif					/* AMIGA || ATARI_ST ...*/

#if HIGHC_386
/*
 * MetaWare's HighC 386 macro putc doesn't handle putc('\n') correctly -
 * sometimes a CR is not written out before the LF.  So, redefine
 * macro putc to actually issue an fputc.
 */
#undef putc
#define putc(c,f) fputc(c,f)
#endif					/* HIGHC_386 */

/*
 * End of operating-system specific code.
 */


#ifndef IncrPtr
#define IncrPtr(p,i) ((p)+(word)(i))
#endif					/* IncrPtr */

#ifndef DecrPtr
#define DecrPtr(p,i) ((p)-(word)(i))
#endif					/* DecrPtr */

#ifndef DiffPtrs
#define DiffPtrs(p1,p2) (word)((p1)-(p2))
#endif					/* Diffptrs */

#ifndef AllocReg
#define AllocReg(n) malloc((msize)n)
#endif					/* AllocReg */
