/*
 * Lexical analysis routines.
 */

#include "../h/config.h"
#include "link.h"
#include "general.h"
#include "opcode.h"
#include <math.h>
#include <ctype.h>

int nlflag = 0;		/* newline last seen */

#ifndef EBCDIC
#define tonum(c)	(isdigit(c) ? (c - '0') : ((c & 037) + 9))
#endif

/*
 * getopc - get an opcode from infile, return the opcode number (via
 * sequential search of opcode table) and point id at the name of the opcode.
 */
 
int getopc(id)
char **id;
   {
   register char *s;
   register struct opentry *p;
   register int test;
 
   s = getstr();
   if (s == NULL)
      return EOF;
   for(test=0;test < NOPCODES; test++) {
       p = &optable[test];
       if( strcmp(p->op_name, s) == 0) {
           *id = p->op_name;
           return (p->op_code);
       }
   }
   *id = s;
   return 0;
   }

/*
 * getid - get an identifier from infile, put it in the identifier
 *  table, and return a pointer to it.
 */
char *getid()
   {
   register char *s;

   s = getstr();
   if (s == NULL)
      return NULL;
   return putident((int)strlen(s)+1);
   }

/*
 * getstr - get an identifier from infile and return a pointer to it.
 */
char *getstr()
   {
   register int c;
   register char *p;

   p = lsfree;
   while ((c = getc(infile)) == ' ' || c == '\t') ;
   if (c == EOF)
      return NULL;
   while (c != ' ' && c != '\t' && c != '\n' && c != ',' && c != EOF) {
      if (p >= lsend)
         quit("out of string space");
      *p++ = c;
      c = getc(infile);
      }
   *p = 0;
   nlflag = (c == '\n');
   return lsfree;
   }
/*
 * getrest - get the rest of the line from infile, put it in the identifier
 *  table, and return a pointer to it.
 */
char *getrest()
   {
   register char *s;
   register int c;
   register char *p;

   /* p = lsfree;
      while ((c = getc(infile)) == ' ' || c == '\t');
   if (c == EOF)
      return NULL; */
   p = lsfree - 1;
   while (c != '\n' && c != EOF) {
      if (p >= lsend)
         quit("out of string space");
      *p++ = c;
      c = getc(infile);
      }
   *p = 0;
   nlflag = (c == '\n');
   return putident((int)strlen(lsfree)+1);
   }

/*
 * getdec - get a decimal integer from infile, and return it.
 */
int getdec()
   {
   register int c, n;
   int sign = 1;

   n = 0;
   while ((c = getc(infile)) == ' ' || c == '\t') ;
   if (c == EOF)
      return 0;
   if (c == '-') {
      sign = -1;
      c = getc(infile);
      }
   while (c >= '0' && c <= '9') {
      n = n * 10 + (c - '0');
      c = getc(infile);
      }
   nlflag = (c == '\n');
   return n*sign;
   }

/*
 * getoct - get an octal number from infile, and return it.
 */
int getoct()
   {
   register int c, n;

   n = 0;
   while ((c = getc(infile)) == ' ' || c == '\t') ;
   if (c == EOF)
      return 0;
   while (c >= '0' && c <= '7') {
      n = (n << 3) | (c - '0');
      c = getc(infile);
      }
   nlflag = (c == '\n');
   return n;
   }

/*
 * getint - get an Icon integer from infile, and return it.
 */
long getint()
   {
   register int c;
   register int r;
   long n;

   n = 0L;
   while ((c = getc(infile)) >= '0' && c <= '9')
      n = n * 10 + (c - '0');
   if (c == 'r' || c == 'R') {
      r = n;
      n = 0L;
      while (c = getc(infile)) {
         if(isdigit(c) || isalpha(c))
            c = tonum(c);
         else
            break;
         n = n * r + c;
         }
      }
   nlflag = (c == '\n');
   return n;
   }

/*
 * getreal - get an Icon real number from infile, and return it.
 */
double getreal()
   {
   double n;
   register int c, d, e;
   int esign;

#ifdef NoAtof
   n = 0.0;
   d = e = 0;
   esign = 1;
   while ((c = getc(infile)) >= '0' && c <= '9')
      n = n * 10 + (c - '0');
   if (c == '.') {
      while ((c = getc(infile)) >= '0' && c <= '9') {
         n = n * 10 + (c - '0');
         d++;
         }
      }
   if (c == 'e' || c == 'E') {
      if ((c = getc(infile)) == '+' || c == '-') {
         if (c == '-')
            esign = -1;
         c = getc(infile);
         }
      while (c >= '0' && c <= '9') {
         e = e * 10 + (c - '0');
         c = getc(infile);
         }
      }
   if (esign < 0)
      e = -e;
   e = e - d;
   if (e <= -LogHuge)
      e = -LogHuge + 1;
   if (e >= LogHuge)
      e = LogHuge - 1;
   if (e != 0)
      n *= pow(10.0, (double)e);
#else					/* NoAtof */
   register char *s, *ep;
   char cbuf[128];
   double atof();

   s = cbuf;
   d = 0;
   while ((c = getc(infile)) == '0')
      ;
   while (c >= '0' && c <= '9') {
      *s++ = c;
      d++;
      c = getc(infile);
      }
   if (c == '.') {
      if (s == cbuf)
         *s++ = '0';
      *s++ = c;
      while ((c = getc(infile)) >= '0' && c <= '9')
         *s++ = c;
      }
   ep = s;
   if (c == 'e' || c == 'E') {
      *s++ = c;
      if ((c = getc(infile)) == '+' || c == '-') {
         esign = (c == '-');
         *s++ = c;
         c = getc(infile);
         }
      e = 0;
      while (c >= '0' && c <= '9') {
         e = e * 10 + c - '0';
         *s++ = c;
         c = getc(infile);
         }
      if (esign) e = -e;
      e += d - 1;
      if (abs(e) >= LogHuge)
         *ep = '\0';
      }
   *s = '\0';
   n = atof(cbuf);
#endif					/* NoAtof */
   nlflag = (c == '\n');
   return n;
   }

/*
 * getlab - get a label ("L" followed by a number) from infile,
 *  and return the number.
 */

int getlab()
   {
   register int c;

   while ((c = getc(infile)) != 'L' && c != EOF && c != '\n') ;
   if (c == 'L')
      return getdec();
   nlflag = (c == '\n');
   return 0;
   }

/*
 * getstrlit - get a string literal from infile, as a string
 *  of octal bytes, and return it.
 */
char *getstrlit(l)
register int l;
   {
   register char *p;

   if (lsfree + l > lsend)
      quit("out of string space");
   p = lsfree;
   while (!nlflag && l--)
      *p++ = getoct();
   *p++ = 0;
   return putident((int)(p-lsfree));
   }

/*
 * newline - skip to next line.
 */
novalue newline()
   {
   register int c;

   if (!nlflag) {
      while ((c = getc(infile)) != '\n' && c != EOF) ;
      }
   nlflag = 0;
   }
