/*
 * File: fmisc.c
 *  Contents: char, collect, copy, display, errorclear, iand, icom, image, ior,
 *  ishift, ixor, ord, runerr, seq, sort, type
 */

#include "../h/config.h"
#include "../h/rt.h"
#include "gc.h"

extern word coll_tot;
extern word coll_stat;
extern word coll_str;
extern word coll_blk;

#ifdef PreProcess
/* include(../m4/fncs.m4) /* */
/* */
#endif					/* PreProcess */

/*
 * char(i) - produce a string consisting of character i.
 */
FncDcl(char,1)
   {
   word i;
   char c;

   if (cvint(&Arg1, &i) == CvtFail)
      RunErr(101, &Arg1);
   if (i < 0 || i >= 256)
      RunErr(205, &Arg1);
   if (strreq((uword)1) == Error)
      RunErr(0, NULL);
   c = i;
   StrLen(Arg0) = 1;
   StrLoc(Arg0) = alcstr(&c, (word)1);
   Return;
   }

/*
 * collect(r,n) - call garbage collector to ensure n bytes in region r.
 */

FncDcl(collect,2)
   {
   long region, bytes;
   word coll = coll_tot;

   if ((defint(&Arg1, &region, (word)0) == Error) ||
       (defint(&Arg2, &bytes, (word)0) == Error)) 
      RunErr(0, NULL);
   if (bytes < 0)
      RunErr(205, &Arg2);
   switch ((int)region) {
      case 0:
         break;
      case Static:
         coll_stat++;
         break;
      case Strings:
         coll_str++;
         strreq((uword)bytes);
         break;
      case Blocks:  
         coll_blk++;
         blkreq((uword)bytes);
         break;
      default:
         RunErr(205, &Arg1);
      };
   if (coll == coll_tot)
      collect((int)region);
   Arg0 = nulldesc;
   Return;
   }

/*
 * copy(x) - make a copy of object x.
 */

FncDcl(copy,1)
   {
   register int i;
   struct descrip  *d1, *d2;
   union block *bp, *ep, **tp;

   if (Qual(Arg1))
      /*
       * Arg1 is a string; just copy its descriptor
       *  into Arg0.
       */
      Arg0 = Arg1;
   else {
      switch (Type(Arg1)) {
         case T_Null:
         case T_Integer:
         case T_Real:
         case T_File:
         case T_Cset:
         case T_Proc:
         case T_Coexpr:
         case T_External:
            /*
             * Copy the null value, integers, long integers, reals, files,
             *	csets, procedures, and such by copying the descriptor.
             *	Note that for integers, this results in the assignment
             *	of a value, for the other types, a pointer is directed to
             *	a data block.
             */
            Arg0 = Arg1;
            break;

         case T_List:
            /*
             * Pass the buck to cplist to copy a list.
             */
            if (cplist(&Arg1, &Arg0, (word)1, BlkLoc(Arg1)->list.size + 1) ==
                 Error) 
               RunErr(0, NULL);
            break;

         case T_Table:
            /*
             * Allocate space for table and elements and copy old table
             *	block into new.
             */
            if (blkreq((sizeof(struct b_table)) +
                  (sizeof(struct b_telem)) * BlkLoc(Arg1)->table.size) ==
                  Error) 
               RunErr(0, NULL);
            bp = (union block *) alctable(&nulldesc);
            bp->table = BlkLoc(Arg1)->table;
            /*
             * Work down the chain of table element blocks in each bucket
             *	and create identical chains in new table.
             */
            for (i = 0; i < TSlots; i++) {
               tp = &(BlkLoc(bp->table.buckets[i]));
               for (ep = *tp; ep != NULL; ep = *tp) {
                  *tp = (union block *)alctelem();
                  (*tp)->telem = ep->telem;
                  tp = &(BlkLoc((*tp)->telem.clink));
                  }
               }
            /*
             * Return the copied table.
             */
            Arg0.dword = D_Table;
            BlkLoc(Arg0) = bp;
            break;

         case T_Set:
            /*
             * Allocate space for set and elements and copy old set
             *	block into new.
             */
            if (blkreq((sizeof(struct b_set)) +
                 (sizeof(struct b_selem)) * BlkLoc(Arg1)->set.size) == Error) 
               RunErr(0, NULL);
            bp = (union block *)alcset();
            bp->set = BlkLoc(Arg1)->set;
            /*
             * Work down the chain of set elements in each bucket
             *	and create identical chains in new set.
             */
            for (i = 0; i < SSlots; i++) {
               tp = &(BlkLoc(bp->set.sbucks[i]));
               for (ep = *tp; ep != NULL; ep = *tp) {
                  *tp = (union block *) alcselem(&nulldesc,(word)0);
                  (*tp)->selem = ep->selem;
                  tp = &(BlkLoc((*tp)->selem.clink));
                  }
               }
            /*
             * Return the copied set.
             */
            Arg0.dword = D_Set;
            BlkLoc(Arg0) = bp;
            break;

         case T_Record:
            /*
             * Allocate space for the new record and copy the old
             *	one into it.
             */
            if (blkreq(BlkLoc(Arg1)->record.blksize) == Error) 
               RunErr(0, NULL);
            i = BlkLoc(BlkLoc(Arg1)->record.recdesc)->proc.nfields;
            bp = (union block *)alcrecd(i,&BlkLoc(Arg1)->record.recdesc);
            bp->record = BlkLoc(Arg1)->record;
            d1 = bp->record.fields;
            d2 = BlkLoc(Arg1)->record.fields;
            while (i--)
               *d1++ = *d2++;
            /*
             * Return the copied record
             */
            Arg0.dword = D_Record;
            BlkLoc(Arg0) = bp;
            break;

         default:
            RunErr(123,&Arg1);
         }
      }
   Return;
   }

/*
 * display(i,f) - display local variables of i most recent
 * procedure activations, plus global variables.
 * Output to file f (default &errout).
 */

FncDcl(display,2)
   {
   long l;
   int count;
   FILE *f;

   /*
    * Arg1 defaults to &level; Arg2 defaults to &errout.
    */
   if ((defint(&Arg1, &l, (word)k_level) == Error) ||
       (deffile(&Arg2, &errout) == Error)) 
      RunErr(0, NULL);

   /*
    * Produce error if file cannot be written.
    */
   f = BlkLoc(Arg2)->file.fd;
   if ((BlkLoc(Arg2)->file.status & Fs_Write) == 0) 
      RunErr(213, &Arg2);

   /*
    * Produce error if Arg1 is negative; constrain Arg1 to be >= &level.
    */
   if (l < 0) 
      RunErr(205, &Arg1)
   else if (l > k_level)
      count = k_level;
   else
      count = l;

   fprintf(f,"co-expression #%ld (%ld)\n\n",BlkLoc(k_current)->coexpr.id,
      BlkLoc(k_current)->coexpr.size);
   fflush(f);
   xdisp(pfp,argp,count,f);
   Arg0 = nulldesc;		/* Return null value. */
   Return;
   }

/*
 * errorclear() - clear error condition.
 */

FncDcl(errorclear,0)
   {
   k_errornumber = 0;
   k_errortext = "";
   k_errorvalue = nulldesc;
   Arg0 = nulldesc;
   Return;
   }

/*
 * iand(i,j) - produce bitwise AND of i and j.
 */
FncDcl(iand,2)
   {
   word i, j;

   if (cvint(&Arg1, &i) == CvtFail)
      RunErr(101, &Arg1);
   if (cvint(&Arg2, &j) == CvtFail)
      RunErr(101, &Arg2);
   MakeInt(i & j, &Arg0);
   Return;
   }

/*
 * icom(i) - produce bitwise complement (one's complement) of i.
 */
FncDcl(icom,1)
   {
   word i;

   if (cvint(&Arg1, &i) == CvtFail)
      RunErr(101, &Arg1);
   MakeInt(~i, &Arg0);
   Return;
   }

/*
 * image(x) - return string image of object x.	Nothing fancy here,
 *  just plug and chug on a case-wise basis.
 */

FncDcl(image,1)
   {
   if (getimage(&Arg1,&Arg0) == Error)
      RunErr(0, NULL);
   Return;
   }

/*
 * ior(i,j) - produce bitwise inclusive OR of i and j.
 */
FncDcl(ior,2)
   {
   word i, j;

   if (cvint(&Arg1, &i) == CvtFail)
      RunErr(101, &Arg1);
   if (cvint(&Arg2, &j) == CvtFail)
      RunErr(101, &Arg2);
   MakeInt(i | j, &Arg0);
   Return;
   }

/*
 * ishift(i,j) - produce i shifted j bit positions (left if j<0, right if j>0).
 */
FncDcl(ishift,2)
   {
   uword i;	/* unsigned to ensure zero fill on right shift */
   word j;

   if (cvint(&Arg1, (word *) &i) == CvtFail)
      RunErr(101, &Arg1);
   if (cvint(&Arg2, &j) == CvtFail)
      RunErr(101, &Arg2);
   /*
    * Check for a shift of WordSize or greater; return an explicit 0 because
    *  this is beyond C's defined behavior.  Otherwise shift as requested.
    */
   if (j <= -WordBits || j >= WordBits)
      i = 0;
   else if (j < 0)
      i >>= -j;
   else
      i <<= j;
   MakeInt(i, &Arg0);
   Return;
   }

/*
 * ixor(i,j) - produce bitwise exclusive OR of i and j.
 */
FncDcl(ixor,2)
   {
   word i, j;

   if (cvint(&Arg1, &i) == CvtFail)
      RunErr(101, &Arg1);
   if (cvint(&Arg2, &j) == CvtFail)
      RunErr(101, &Arg2);
   MakeInt(i ^ j, &Arg0);
   Return;
   }

/*
 * ord(s) - produce integer ordinal (value) of single chracter.
 */
FncDcl(ord,1)
   {
   char sbuf[MaxCvtLen];

   if (cvstr(&Arg1, sbuf) == CvtFail)
      RunErr(103, &Arg1);
   if (StrLen(Arg1) != 1)
      RunErr(205, &Arg1);
   MakeInt(*StrLoc(Arg1) & 0xFF, &Arg0);
   Return;
   }

/*
 * runerr(i,x) - produce runtime error i with value x.
 */

FncDclV(runerr)
   {
   long l;

   if (nargs < 1)
      RunErr(-101, NULL);

   switch (cvint(&Arg(1), &l)) {
       case T_Integer:
           if (l <= 0)
              RunErr(205, &Arg(1));
	   break;

       default:
          RunErr(101, &Arg(1));
       }

   if (nargs == 1) {
      RunErr((int)(-l), NULL);
      }
   else {
      RunErr((int)l, &Arg(2));
      }
      
   }

/*
 * seq(e1,e2) - generate e1, e1+e2, e1+e2+e2, ... .
 */

FncDcl(seq,2)
   {
   long from, by;

   /*
    * Default Arg1 and Arg2 to 1.
    */
   if ((defint(&Arg1, &from, (word)1) == Error) ||
       (defint(&Arg2, &by, (word)1) == Error)) 
      RunErr(0, NULL);
   
   /*
    * Produce error if Arg2 is 0, i.e., an infinite sequence of Arg2s.
    */
   if (by == 0) 
      RunErr(211, &Arg2);

   /*
    * Suspend sequence, stopping when largest or smallest integer
    *  is reached.
    */
   while ((from <= MaxLong && by > 0) || (from >= MinLong && by < 0)) {
      MakeInt(from, &Arg0);
      Suspend;
      from += by;
      }
   Fail;
   }

struct dpair {
   struct descrip dr;
   struct descrip dv;
   };

/*
 * sort(l) - sort list l.
 * sort(S) - sort set S.
 * sort(t,i) - sort table.
 */

FncDcl(sort,2)
   {
   register dptr d1;
   register word size;
   register int i;
   word nslots;
   struct b_list *lp, *tp;
   union block *bp, *ep;

   if (Arg1.dword == D_List) {
      /*
       * Sort the list by copying it into a new list and then using
       *  qsort to sort the descriptors.  (That was easy!)
       */
      size = BlkLoc(Arg1)->list.size;
      if (cplist(&Arg1, &Arg0, (word)1, size + 1) == Error) 
         RunErr(0, NULL);
      qsort((char *)BlkLoc(BlkLoc(Arg0)->list.listhead)->lelem.lslots,
         (int)size, sizeof(struct descrip), anycmp);
      }
   else if (Arg1.dword == D_Set) {
      /*
       * Create a list the size of the set (or at least 
       *  MinListSlots), copy each element into the list, and
       *  then sort the list using qsort as in list sorting
       *  and return the sorted list.
       */
   nslots = size = BlkLoc(Arg1)->set.size;
   if(nslots < MinListSlots)
      nslots = MinListSlots;

   if (blkreq(sizeof(struct b_list) + sizeof(struct b_lelem) +
      nslots * sizeof(struct descrip)) == Error) 
      RunErr(0, NULL);

   bp = BlkLoc(Arg1);
   lp = alclist(size);
   lp->listhead.dword = lp->listtail.dword = D_Lelem;
      BlkLoc(lp->listtail) = (union block *)alclstb(nslots, (word)0, size);
   BlkLoc(lp->listhead) = BlkLoc(lp->listtail);
   if (size > 0) {  /* only need to sort non-empty sets */
      d1 = BlkLoc(lp->listhead)->lelem.lslots;
      for(i = 0; i < SSlots; i++) {
      ep = BlkLoc(bp->set.sbucks[i]);
      while (ep != NULL) {
         *d1 = ep->selem.setmem;
         d1++;
         ep = BlkLoc(ep->selem.clink);
         }
      }
      qsort((char *)BlkLoc(lp->listhead)->lelem.lslots,(int)size,
         sizeof(struct descrip),anycmp);
   }
   Arg0.dword = D_List;
   BlkLoc(Arg0) = (union block *) lp;
   }

   else if (Arg1.dword == D_Table) {
      /*
       * Default i (the type of sort) to 1.
       */
      if (defshort(&Arg2, 1) == Error) 
         RunErr(0, NULL);
      switch ((int)IntVal(Arg2)) {

      /*
       * Cases 1 and 2 are as in standard Version 5.
       */
         case 1:
         case 2:
		{
      /*
       * The list resulting from the sort will have as many elements as
       *  the table has, so get that value and also make a valid list
       *  block size out of it.
       */
      nslots = size = BlkLoc(Arg1)->table.size;
      if (nslots < MinListSlots)
         nslots = MinListSlots;
      /*
       * Ensure space for: the list header block and a list element
       *  block for the list which is to be returned,
       *  a list header block and a list element block for each of the two
       *  element lists the sorted list is to contain.	Note that the
       *  calculation might be better expressed as:
       *    list_header_size + list_block_size + nslots * descriptor_size +
       *     nslots * (list_header_size + list_block_size + 2*descriptor_size)
       */
      if (blkreq(sizeof(struct b_list) + sizeof(struct b_lelem) +
         nslots * (sizeof(struct b_list) + sizeof(struct b_lelem) +
            3 * sizeof(struct descrip))) == Error) 
         RunErr(0, NULL);
      /*
       * Point bp at the table header block of the table to be sorted
       *  and point lp at a newly allocated list
       *  that will hold the the result of sorting the table.
       */
      bp = BlkLoc(Arg1);
      lp = alclist(size);
      lp->listhead.dword = lp->listtail.dword = D_Lelem;
      BlkLoc(lp->listtail) = (union block *)alclstb(nslots, (word)0, size);
      BlkLoc(lp->listhead) = BlkLoc(lp->listtail);
      /*
       * If the table is empty, there is no need to sort anything.
       */
      if (size <= 0)
         break;
         /*
          * Point d1 at the start of the list elements in the new list
          *  element block in preparation for use as an index into the list.
          */
         d1 = BlkLoc(lp->listhead)->lelem.lslots;
         /*
          * Traverse the element chain for each table bucket.  For each
          *  element, allocate a two-element list and put the table
          *  entry value in the first element and the assigned value in
          *  the second element.  The two-element list is assigned to
          *  the descriptor that d1 points at.	When this is done, the
          *  list of two-element lists is complete, but unsorted.
          */
         for (i = 0; i < TSlots; i++) {
            ep = BlkLoc(bp->table.buckets[i]);
            while (ep != NULL) {
               d1->dword = D_List;
               tp = alclist((word)2);
               BlkLoc(*d1) = (union block *) tp;
               tp->listhead.dword = tp->listtail.dword = D_Lelem;
               BlkLoc(tp->listtail) = (union block *)alclstb((word)2, (word)0,
                  (word)2);
               BlkLoc(tp->listhead) = BlkLoc(tp->listtail);
               BlkLoc(tp->listhead)->lelem.lslots[0] = ep->telem.tref;
               BlkLoc(tp->listhead)->lelem.lslots[1] = ep->telem.tval;
               d1++;
               ep = BlkLoc(ep->telem.clink);
               }
            }
         /*
          * Sort the resulting two-element list using the sorting function
          *  determined by i.
          */
         if (IntVal(Arg2) == 1)
            qsort((char *)BlkLoc(lp->listhead)->lelem.lslots, (int)size,
                  sizeof(struct descrip), trefcmp);
         else
            qsort((char *)BlkLoc(lp->listhead)->lelem.lslots, (int)size,
                  sizeof(struct descrip), tvalcmp);
         break;		/* from cases 1 and 2 */
         }
      /*
       * Cases 3 and 4 were introduced in Version 5.10.
       */
         case 3 :
         case 4 :
                 {
      /*
       * The list resulting from the sort will have twice as many elements as
       *  the table has, so get that value and also make a valid list
       *  block size out of it.
       */
      nslots = size = BlkLoc(Arg1)->table.size * 2;
      if (nslots < MinListSlots)
         nslots = MinListSlots;
      /*
       * Ensure space for: the list header block and a list element
       *  block for the list which is to be returned, and two descriptors for
       *  each table element.
       */
      if (blkreq(sizeof(struct b_list) + Vsizeof(struct b_lelem) +
            (nslots * sizeof(struct descrip))) == Error) 
         RunErr(0, NULL);

      /*
       * Point bp at the table header block of the table to be sorted
       *  and point lp at a newly allocated list
       *  that will hold the the result of sorting the table.
       */
      bp = BlkLoc(Arg1);
      lp = alclist(size);
      lp->listhead.dword = lp->listtail.dword = D_Lelem;
      BlkLoc(lp->listtail) = (union block *)alclstb(nslots, (word)0, size);
      BlkLoc(lp->listhead) = BlkLoc(lp->listtail);
      /*
       * If the table is empty there's no need to sort anything.
       */
      if (size <= 0)
         break;

         /*
          * Point d1 at the start of the list elements in the new list
          *  element block in preparation for use as an index into the list.
          */
         d1 = BlkLoc(lp->listhead)->lelem.lslots;
         /*
          * Traverse the element chain for each table bucket.  For each
          *  table element copy the the entry descriptor and the value
          *  descriptor into adjacent descriptors in the lslots array
          *  in the list element block.
          *  When this is done we now need to sort this list.
          */
         for (i = 0; i < TSlots; i++) {
            ep = BlkLoc(bp->table.buckets[i]);
            while (ep != NULL) {
               *d1 = ep->telem.tref;
               d1++;
               *d1 = ep->telem.tval;
               d1++;
               ep = BlkLoc(ep->telem.clink);
               }
            }
         /*
          * Sort the resulting two-element list using the sorting function
          *  determined by i.
          */
         if (IntVal(Arg2) == 3)
            qsort((char *)BlkLoc(lp->listhead)->lelem.lslots, (int)size / 2,
                  (2 * sizeof(struct descrip)), trcmp3);
         else
            qsort((char *)BlkLoc(lp->listhead)->lelem.lslots, (int)size / 2,
                  (2 * sizeof(struct descrip)), tvcmp4);
            break; /* from case 3 or 4 */
            }

         default:
            RunErr(205, &Arg2);

         } /* end of switch statement */

      /*
       * Make Arg0 point at the sorted list.
       */
      Arg0.dword = D_List;
      BlkLoc(Arg0) = (union block *) lp;
      }
   else {  /* Tried to sort something that wasn't a list or a table. */
      RunErr(115, &Arg1);
      }
   Return;
   }

/*
 * trefcmp(d1,d2) - compare two-element lists on first field.
 */

trefcmp(d1,d2)
dptr d1, d2;
   {

#ifdef DeBug
   if (d1->dword != D_List || d2->dword != D_List)
      syserr("trefcmp: internal consistency check fails.");
#endif					/* DeBug */

   return (anycmp(&(BlkLoc(BlkLoc(*d1)->list.listhead)->lelem.lslots[0]),
                  &(BlkLoc(BlkLoc(*d2)->list.listhead)->lelem.lslots[0])));
   }

/*
 * tvalcmp(d1,d2) - compare two-element lists on second field.
 */

tvalcmp(d1,d2)
dptr d1, d2;
   {

#ifdef DeBug
   if (d1->dword != D_List || d2->dword != D_List)
      syserr("tvalcmp: internal consistency check fails.");
#endif					/* DeBug */

   return (anycmp(&(BlkLoc(BlkLoc(*d1)->list.listhead)->lelem.lslots[1]),
                  &(BlkLoc(BlkLoc(*d2)->list.listhead)->lelem.lslots[1])));
   }

/*
 * The following two routines are used to compare descriptor pairs in the
 *  experimental table sort.
 *
 * trcmp3(dp1,dp2)
 */

trcmp3(dp1,dp2)
struct dpair *dp1,*dp2;
{
   return (anycmp(&((*dp1).dr),&((*dp2).dr)));
}
/*
 * tvcmp4(dp1,dp2)
 */

tvcmp4(dp1,dp2)
struct dpair *dp1,*dp2;

   {
   return (anycmp(&((*dp1).dv),&((*dp2).dv)));
   }

/*
 * type(x) - return type of x as a string.
 */

FncDcl(type,1)
   {

   if (Qual(Arg1)) {
      StrLen(Arg0) = 6;
      StrLoc(Arg0) = "string";
      }

   else {
      switch (Type(Arg1)) {

         case T_Null:
            StrLen(Arg0) = 4;
            StrLoc(Arg0) = "null";
            break;

         case T_Integer:
            StrLen(Arg0) = 7;
            StrLoc(Arg0) = "integer";
            break;

         case T_Real:
            StrLen(Arg0) = 4;
            StrLoc(Arg0) = "real";
            break;

         case T_Cset:
            StrLen(Arg0) = 4;
            StrLoc(Arg0) = "cset";
            break;

         case T_File:
            StrLen(Arg0) = 4;
            StrLoc(Arg0) = "file";
            break;

         case T_Proc:
            StrLen(Arg0) = 9;
            StrLoc(Arg0) = "procedure";
            break;

         case T_List:
            StrLen(Arg0) = 4;
            StrLoc(Arg0) = "list";
            break;

         case T_Table:
            StrLen(Arg0) = 5;
            StrLoc(Arg0) = "table";
            break;

         case T_Set:
            StrLen(Arg0) = 3;
            StrLoc(Arg0) = "set";
            break;

         case T_Record:
            Arg0 = BlkLoc(BlkLoc(Arg1)->record.recdesc)->proc.recname;
            break;

         case T_Coexpr:
            StrLen(Arg0) = 13;
            StrLoc(Arg0) = "co-expression";
            break;

         case T_External:
            StrLen(Arg0) = 8;
            StrLoc(Arg0) = "external";
            break;

         default:
            RunErr(123,&Arg1);
         }
      }
   Return;
   }
