/*
 * File: fxtra.c
 *  Contents: additional functions to extend the standard Icon repertoire.
 *  This files contains several collections of functions:
 *
 *	Mathematical functions (MathFncs)
 *	Functions specific to MS-DOS (DosFncs)
 *	Memory monitor functions (MemMon)
 *	Instrumentation functions (InstrFncs)
 *	Internals functions (IntrnlFncs)
 *	"Iscope" functions (ScopeFncs)
 *	Screen Functions (ScreenFncs)
 *	String Functions (StringFncs)
 *	Miscellaneous functions (MiscFncs)
 *
 *  These collections are under the control of conditional compilation
 *  as indicated by the symbols in parentheses. To enable a set of functions,
 *  define the corresponding symbol in ../h/define.h.  The functions themselves
 *  are in separate files, included according to the defined symbols.
 *
 *  CAVEAT:  Most of these functions have not yet been tested.  In
 *           addition, there are operating-system dependencies that
 *           are not noted.
 *
 *  The following persons contributed functions: Bob Alexander,Ralph Griswold,
 *  Andy Heron, Bill Mitchell, Gregg Townsend, Steve Wampler, and Cheyenne
 *  Wills.
 */

#include "../h/config.h"
#include "../h/rt.h"
#include "gc.h"

#ifdef MathFncs
#include "fxmath.c"
#endif					/* MathFncs */

#ifdef DosFncs
#include "fxmsdos.c"
#endif					/* DosFncs */

#ifdef MemMon
#include "fxmemmon.c"
#endif					/* MemMon */

#ifdef InstrFncs
#include "fxinstr.c"
#endif					/* InstrFncs */

#ifdef IntrnlFncs
#include "fxintrnl.c"
#endif					/* IntrnlFncs */

#ifdef ScopeFncs
#include "fxiscope.c"
#endif					/* ScopeFncs */

#ifdef ScreenFncs
#include "fxscreen.c"
#endif					/* ScreenFncs */

#ifdef StringFncs
#include "fxstr.c"
#endif					/* StringFncs */

#ifdef DebugFncs
#endif					/* DebugFncs */

#ifdef MiscFncs
#include "fxsys.c"
#endif					/* MiscFncs */

#if VM
#include "fxvm.c"
#endif					/* VM */

char fx_junk;				/* prevent empty module */
