/*
 * Definitions and declarations used for storage management.
 */

#define F_Mark		0100000 	/* bit for marking blocks */

#define GranSize	     64		/* storage allocation granule size */

#define Sqlinc		    128 	/* qualifier pointer list increment */

#define Static  1			/* collection is for static region */
#define Strings	2			/* collection is for strings */
#define Blocks	3			/* collection is for blocks */

/*
 * External definitions.
 */

extern char *currend;			/* current end of memory region */
extern uword blkneed;			/* stated need for block space */
extern uword strneed;			/* stated need for string space */
extern uword statneed;
extern dptr globals; 			/* start of global variables */
extern dptr eglobals;			/* end of global variables */
extern dptr gnames;			/* start of global variable names */
extern dptr egnames; 			/* end of global variable names */
extern dptr statics; 			/* start of static variables */
extern dptr estatics;			/* end of static variables */

extern dptr *quallist;			/* start of qualifier list */
extern word qualsize;

/*
 * Get type of block pointed at by x.
 */
#define BlkType(x)   (*(word *)x)

/*
 * BlkSize(x) takes the block pointed to by x and if the size of
 *  the block as indicated by bsizes[] is nonzero it returns the
 *  indicated size; otherwise it returns the second word in the
 *  block contains the size.
 */
#define BlkSize(x) (bsizes[*(word *)x & ~F_Mark] ? \
		     bsizes[*(word *)x & ~F_Mark] : *((word *)x + 1))

/*
 * If memory monitoring is not enabled, redefine function calls
 * to do nothing.
 */
#ifndef MemMon
#define MMAlc(n,t)
#define MMBGC(r)
#define MMEGC()
#define MMMark(b,t)
#define MMStat(a,l,c)
#define MMStr(n)
#define MMSMark(a,n)
#endif					/* MemMon */

#ifndef FixedRegions

/*
 * Information used with Icon's allocation routines with expandable-regions
 *  memory management.
 */

typedef int ALIGN;		/* pick most stringent type for alignment */

union bhead {			/* header of free block */
   struct {
      union bhead *ptr; 	/* pointer to next free block */
      uword bsize;		/* free block size */
      } s;
   ALIGN x;			/* force block alignment */
   };

typedef union bhead HEADER;
#define NALLOC 64		/* units to request at one time */

#define FREEMAGIC 0x807F	/* magic flag for free blocks (MemMon only) */

#endif					/* FixedRegions */
