/*
 * Routines needed for different systems.
 */

#include "../h/config.h"
#include "../h/rt.h"
#include <ctype.h>
#include <math.h>

/*
 * The following code is operating-system dependent [@rlocal.01].
 *  Routines needed by different systems.
 */

#if PORT
   /* place for anything system-specific */
Deliberate Syntax Error
#endif					/* PORT */

#if AMIGA
/*
 * amigenv.c
 *		The routines in this file provide UNIX-like environment
 *	support for the AMIGA. Specifically, equivalents to the routines
 *	getenv(), and putenv() are provided. The environment is actually
 *	maintained on a disk file. For a more complete description of
 *	the exact system see the documentation file amigenv.doc.
 */
#include <string.h>
#define EOS '\0'

char _EnvFileName[128] = "ENV:";
char _CurrEnvFile[128] = "";
char *_EnvBase;
char *_EnvEnd;
char *_EnvMax;
char **_envp;

char *getenv(envstr)
char *envstr;
{
char **ep;
char *p;
extern char *envfind();
extern char *openenv();

if (envstr == NULL)
	return (openenv());
if (strcmp(_CurrEnvFile,_EnvFileName) != 0)
	if (openenv() == NULL)
		return (NULL);
for (ep = _envp; *ep != NULL; ep++)
	if (p = envfind(envstr,*ep))
		return p;
return NULL;
}

char *openenv()
{
FILE *efp;
char *p;
int n;
char buf[128];
extern char *normalize();
extern char *_EnvBase;
if ((efp = fopen(_EnvFileName,"r")) == NULL) {
	printf("can't open environment file\n");
	exit(1);
	}
if (_EnvBase == NULL) {
	_EnvBase = calloc(500,1);
	_EnvMax = _EnvBase + 500;
	_envp = (char **)calloc(25,sizeof(int));
	}
p = _EnvBase;
while (fgets(buf,128,efp)) {
	if (normalize(buf) == NULL)
		continue;
	p = stpcpy(p,buf);
	p--;
	*p = EOS;
	p++;
	if (p >= _EnvMax) {
		fprintf(stderr,"syserr: env overflow in openenv()\n");
		exit(1);
		}
	}
*p = EOS;
_EnvEnd = p;
fclose(efp);
if ((n = strbpl(_envp,25,_EnvBase)) < 0 )
	return ((char *)NULL); /* failure */
else {
	strcpy(_CurrEnvFile,_EnvFileName);
	return (_EnvBase); /* ok */
	}
}

char *normalize(str)
char *str;
{
char *s, *d;

if (strchr(str,'=') == NULL)
	return ((char *)NULL);


rpt:
d = str;
if ((d = strpbrk(d," \t")) != NULL) {
	if ((s = stpblk(d)) == NULL) {
		*d = EOS;
		return (strchr(str,'='));
		}
	while ((*d++ = *s++) != NULL)
		continue;
	goto rpt;
	}
else
	return (strchr(str,'='));
}

int putenv(envpair)
char *envpair;
{
char varbuf[128];
char *p, *s, *t;
char **ep;
FILE *efp;
extern FILE *fopen();
extern char *normalize();

strcpy(varbuf,envpair);
if ((p = normalize(varbuf)) == NULL)
	return 0;
*p = EOS;
for (ep = _envp; *ep != NULL; ep++) {
	s = varbuf;
	t = *ep;
	while (*s == *t) {
		s++;
		t++;
	  }
	if (*s != EOS || *t != '=')
		continue;
	else
		break;
	}
if (*ep == NULL) {
	*ep = _EnvEnd;
	ep++;
	*ep = NULL;
	}
else
	*ep = _EnvEnd;

*p = '=';
p = stpcpy(_EnvEnd,varbuf);
p++;
*p = EOS;
_EnvEnd = p;
if ((efp = fopen(_EnvFileName,"w")) == NULL) {
	fprintf(stderr,"can't open environment file for writing\n");
	exit(1);
	}
for (ep = _envp; *ep != NULL; ep++) {
	fputs(*ep,efp);
	fputc('\n',efp);
	}
fclose(efp);
strcpy(_CurrEnvFile,_EnvFileName);
return 1;
}
/*
 * envfind - returns pointer to value string in environment pairs
 *           returns NULL if the variable is not found.
 */

char *envfind(str,estr)
   char *str,*estr;
{
char *p,*q;

p = str;
q = estr;
while (*p == *q) {
	p++;
	q++;
}
if (*p != EOS || *q != '=')
	return ((char *)NULL);
else
	return (++q);
}
novalue envassign(pathname)
char * pathname;
{

strcpy(_EnvFileName,pathname);
return;
}

novalue envshow()
{
char **ep;
int n;
extern char *openenv();
extern char **_envp;
if (_EnvBase == NULL)
	openenv();
if (strcmp(_EnvFileName,_CurrEnvFile) != 0)
	openenv();
n = 1;
for (ep = _envp; n < 15; ep++) {
	if (*ep == NULL)
		break;
	printf("env #%d ->  %s \n",n,*ep);
	n++;
	}
}
#endif					/* AMIGA */

#if ATARI_ST

#include <dos.h>
#include <string.h>

char iconpath[80];		/* the path returned by findpath() */
int pathset = 0;		/* not yet setup */

char *findpath();
char *getpathenv();

/*
 Pass a command to a shell.
 Find what/where is the shell by the environment variable SHELL
 */
int system(cmd)
char *cmd;
{
	char *shell, tail[128];
	int rc;
	
	if ((shell = getenv("SHELL")) == NULL) {
		fprintf(stderr,"SHELL not found.\n");
		return -1;
	}

	strcpy(tail + 1, cmd);
	*tail = strlen(tail) - 1;

	rc = Pexec(shell, tail, NULL);
	return (rc >= 0) ? rc : -1;
}

/*
 Spawn a process,
 with: variable arg list, and path found by environment.
 */
int Forkvp(cmd, argv)
char *cmd, **argv;
{
	char file[80];
	int n;
	char *patharg = NULL;

	if (filefound(NULL, cmd) == 0) {
		if (!pathset) {
			if (patharg) {
				strcpy(iconpath, patharg);
				if ((n = strlen(iconpath)) > 0
						&& iconpath[n-1] != '\\')
					strcat(iconpath, "\\");
			} else {
				if (findpath(cmd) == NULL)
					return -1;
			}
			pathset = 1;
		}
		strcpy(file, iconpath);
		strcat(file, cmd);
	} else
		strcpy(file, cmd);

	return Forkv(file, argv);	
}

/*
 Spawn a process,
 with: variable arg list.
 */
int Forkv(cmd, argv)
char *cmd, **argv;
{
	char tail[128];
	int i;

	*tail = tail[1] = '\0';
	for (i = 1; argv[i]; i++) {
		strcat(tail," ");
		strcat(tail,argv[i]);
	}
	*tail = strlen(tail) - 1;

	return Pexec(cmd, tail, NULL);	
}

/*
 findpath() assumes that the name passed to it is complete.
 (i.e. with the proper extension...)
 */  
char *findpath(prog)
char *prog;
{
	char *ppath, *fpath;
	int error, drvnum, drvmap, drvmask;
/*	
 First : check if ICON environment variable exist...
 (because it may be a lot faster than check the whole PATH...)
 */
	if ((ppath = getenv("ICON")) != NULL) {
		strcpy(iconpath, ppath);
		if ((error = strlen(iconpath)) > 0) {
			if (iconpath[error - 1] != '\\')
				strcat(iconpath, "\\");
			if (filefound(iconpath, prog) != 0)
				return iconpath;
		}
	}
/*	
 Second : check with PATH...
 */

	if ((ppath = getenv("PATH")) != NULL) {
		error = 0;
		while ((fpath = getpathenv(ppath, error)) != NULL) {
			if (filefound(fpath, prog) != 0) {
				strcpy(iconpath, fpath);
				return iconpath;
			}
			error++;
		}
	}

/*
 Not found here or with the ICON or PATH environment variable.
 Check all the mounted disk drives root directory, in alphabetical order.
 */
	strcpy(iconpath, "a:\\");
	drvmap = bios(10);	/* Drvmap() */ 
	for (drvnum = 0,	drvmask = 1;
	     drvnum < 16;
	     drvnum++,		drvmask <<= 1, iconpath[0]++)
		if ((drvmask & drvmap) && (filefound(iconpath, prog) != 0))
				return iconpath;

/*
 Not found...
 */
	return NULL;
}

int Pexec(file, tail, env)
char *file, *tail, *env;
{
	int rc;

	if ((rc = gemdos(0x4b,0,file,tail, env)) != 0) {
		switch (rc) {
			case -33:
				fprintf(stderr,
				"Fork failed: file <%s> not found.\n", file);
				break;
			case -39:
				fprintf(stderr,
				"Fork failed: <%s> Not enough memory.\n", file);
				break;
			case -66:
				fprintf(stderr,
				"Fork failed: <%s> Bad load format.\n", file);
				break;
			default:
		}
	}
	return rc;
}

/*
 getpathenv() find the nth path entry in the env var PATH.
 append a backslash to it, if necessary; strcat ready.
 */
 
char *getpathenv(path, n)
char *path; int n;
{
	char *Path2;
	char *sep = " ,;";		/* path separator in PATH */
	static char local[128];
	int i, j;

	for (i = 0; i < n; i++) {
		Path2 = strpbrk(path, sep);
		if (Path2 == NULL)
			break;
		path = Path2 + 1;	/* Past the sep */
		if ((j = strspn(path, sep)) > 0)
			path += j;
		
	}
	if (n == i) {
		strcpy(local, path);
		Path2 = strpbrk(local, sep);
		if (Path2 != NULL) {
			*Path2++ = '\\';
			*Path2 = '\0';
		} else
			strcat(local, "\\");
		if (strlen(local) == 1)
			local[0] = '\0';
		return &local[0];
	}
	return NULL;
}
	
/*
 test is a file exist.
 */
int filefound(path, file)
char *path, *file;
{
	char filepath[80];

	if ((path != NULL) && (path[0] != '\0')) {
		strcpy(filepath, path);
		strcat(filepath, file);
	} else
		strcpy(filepath, file);

	if (gemdos(0x43, filepath, 0, 0) >= 0)
		return 1;
	return 0;
}
#endif					/* ATARI_ST */

#if HIGHC_386
#endif					/* HIGHC_386 */

#if MACINTOSH
#if MPW
/*
**  Special routines for Macintosh Programmer's Workshop
**  implementation of the Icon Programming Language
*/

#include <Types.h>
#include <Events.h>
#include <OSUtils.h>
#include <Memory.h>
#include <Errors.h>
#include "time.h"

#undef power

pascal void MacsBugPrint(s)
  char *s;
  extern 0xABFF;

/*
**  Initialization and Termination Routines
*/
/*
**  MacExit -- This function is installed by an onexit() call in MacInit
**  -- it is called automatically when the program terminates.
*/
void
MacExit()
{
  void ResetStack();
  extern Ptr MemBlock;

  ResetStack();
  if (MemBlock != NULL) DisposPtr(MemBlock);
}

/*
**  MacInit -- This function is called near the beginning of execution of
**  iconx.  It is called by our own brk/sbrk initialization routine.
*/
void
MacInit()
{
  onexit(MacExit);
}


/*
**  UNIX-like Date and Time Routines for Macintosh
*/

long
time(tloc)
long *tloc;
{
  long x;

  GetDateTime(&x);
  if (tloc) *tloc = x;
  return x;
}

struct tm *
localtime(clock)
long *clock;
{
  static struct tm tm;
  struct DateTimeRec dtr;
  static short days[] = {
	-1, 30, 58, 89, 119, 150, 180, 211, 242, 272, 303, 333
  };

  Secs2Date(*clock, &dtr);
  tm.tm_sec = dtr.second;
  tm.tm_min = dtr.minute;
  tm.tm_hour = dtr.hour;
  tm.tm_mday = dtr.day;
  tm.tm_mon = dtr.month - 1;		/* 0-11 */
  tm.tm_year = dtr.year - 1900;		/* year since 1900 */
  tm.tm_wday = dtr.dayOfWeek - 1;	/* 0 = Sunday */
  tm.tm_yday = days[tm.tm_mon] + tm.tm_mday +
	(tm.tm_mon > 1 && (tm.tm_year&3) == 0);
  tm.tm_isdst = 0;	/* Not available on Macintosh */
  tm.tm_hsec = 0;	/* Not available on Macintosh */
  return(&tm);
}


/*
**  Brk and Sbrk Equivalents
*/

typedef Ptr caddr_t;

static caddr_t MemBlock, Break, Limit;
word xcodesize;

init_brk()
{
  static short init = 0;
  Size max, grow, size;
  char *v;
  extern word mstksize, statsize, ssize, abrsize;

  if (!init) {
    init = 1;
    MacInit();
    if ((v = getenv("ICONSIZE")) != NULL) {
      if ((size = atol(v)) <= 0) {
	max = MaxMem(&grow);
	size = max + grow - (size < 0 ? -size : max / 4);
      }
    }
    else { /* ICONSIZE undefined */
      size = xcodesize + mstksize + statsize + ssize + abrsize + 512;
    }
    if ((MemBlock = NewPtr(size)) == NULL) {
      syserr("problem allocating Mac memory");
    }
    Break = MemBlock;
    Limit = MemBlock + size;
  }
  return 1;
}

caddr_t
brk(addr)
caddr_t addr;
{
  Size newsize;

  if (!init_brk()) return (caddr_t)-1;
  if (addr < MemBlock) return (caddr_t)-1;
  if (addr < Limit) Break = addr;
  else {
    newsize = addr - MemBlock;
    SetPtrSize(MemBlock, newsize);
    if (MemError() != noErr) return (caddr_t)-1;
    Break = Limit = addr;
  }
  return (caddr_t)0;
}

caddr_t
sbrk(incr)
int incr;
{
  caddr_t start;

  if (!init_brk()) return (caddr_t)-1;
  start = Break;
  if (incr != 0) {
    if (brk(start + incr) == (caddr_t)-1) return (caddr_t)-1;
  }
  return start;
}


/*
**  Abort()
*/

novalue abort()
{
  MacsBugPrint("\034ICON ERROR WITH ICONCORE SET");
  exit(1);
}
/*
**  Pow() -- to resolve name conflict with Apple SANE routine named
**  "power".
*/

#undef pow
extended pow(x, y)
extended x, y;
{
  extern extended power();

  return power(x, y);
}

/*
**  "Stub" functions to satisfy references to functions that are never
**  called in MPW tools.  (Copied from file {MPW}CExamples:Stubs.c).
*/

/* Console Driver

   These drivers provide I/O to the screen (or a specified port) in
   applications.  They aren't necessary in tools.  
*/

_coFAccess() {}
_coClose() {}
_coRead() {}
_coWrite() {}
_coIoctl() {}
_coExit() {}


/* File System Driver

   Tools use the file system drivers linked with the MPW Shell.
*/

_fsFAccess() {}
_fsClose() {}
_fsRead() {}
_fsWrite() {}
_fsIoctl() {}
_fsExit() {}


/* System Driver

   Tools use the system drivers linked with the MPW Shell.
*/

_syFAccess() {}
_syClose() {}
_syRead() {}
_syWrite() {}
_syIoctl() {}
_syExit() {}
#endif					/* MPW */
#endif					/* MACINTOSH */

#if MSDOS

#if TURBO
extern unsigned _stklen = 8192;
#endif					/* TURBO */

#if LATTICE

#include <error.h>

int _stack = (8 * 1024);
long int _mneed = (20 * 1024);

extern long int *sp;
long int **xsp = &sp;  /* Used for rswitch.asm .. since 'sp' is a reserved */
		       /* symbol for the assembler.. */

extern char *statend;  /* Indicator for when to use malloc for _GETBF */

int brk(p)
char *p;
{
   char *sbrk();
   long int l, m;

   l = (long int)p;
   m = (long int)sbrk((word)0);

   if( lsbrk((long) (l - m) ) == 0) return -1;
   else return 0;
}

/* Rewrite of _GETBF which calls malloc instead of GETMEM after
 * icon has initialized its memory management.
 * Note that we flag any buffers gotten by malloc as private buffers
 */

int _GETBF(fd)
FILE *fd;
{

/* check to see if buffer already allocated (or setbufed) */
   if( fd->_size != 0 || fd->_flag & _IOMYBUF ) return 0;

/* check to see if icon's memory manager is running the show */
   if( statend )   /* Yes - then use malloc */
       fd->_base = fd->_ptr = malloc(_BUFSIZ);
   else 	  /* No - then use getmem */
       fd->_base = fd->_ptr = getmem(_BUFSIZ);

/* Check to see if allocation failed.. */
   if (fd->_base == NULL) {
       errno = ENOMEM; /* No memory available */
       return EOSERR;  /* Return error indicator */
       }
/* Set size to _BUFSIZE */
   fd->_size = _BUFSIZ;

/* Turn off non buffered flag, and private buffer flag */
   fd->_flag &= ~(_IONBF | _IOMYBUF);

/* If buffer malloced flag as private storage */
   if ( statend ) fd->_flag |= _IOMYBUF;

/* Set write and read counters to 0 */
   fd->_wcnt = fd->_rcnt = 0;
   return 0;
}
novalue abort()    /* Abort set to 'dump' icon data area.. */
{
#ifdef DeBug
   blkdump();
#endif					/* DeBug */
   fflush(stderr);
   fcloseall();
   _exit(1);
}
#endif					/* LATTICE */
#endif					/* MSDOS */

#if MVS || VM
const int _staksize = (64*1024);
#endif					/* MVS || VM */

#if OS2
#endif					/* OS2 */

#if UNIX
#ifdef ATTM32
extern int mstksize;

/*
 * This file contains the routine necessary to allocate legal AT&T
 * 3B2/15/4000 stack space for co-expression stacks.
 *
 * Legal stack region begins at 0xC0020000, and UNIX will grow stack space
 * up to 50 Megabytes. 0xC0030000 should provide plenty of room for
 * main C stack growth.  Each time coexpr_salloc() is called, it
 * adds mstksize (max main stack size) and returns a new address,
 * meaning each coexpression stack is potentially as large as the main stack.
 */

/*
 * coexp_salloc() - return pointer in legal stack space for start
 *                  of a coexpression stack.
 */

pointer coexp_salloc()
   {
   static pointer sp = 0xC0030000 ;     /* pointer to stack region */

   sp +=  mstksize;
   return sp;
}
#endif					/* ATTM32 */

#ifdef ExecImages

#ifdef CONVEX
#define TEXT0 0x80001000		/* address of first .text page */
#define DATA0 ((int) &environ & -4096)	/* address of first .data page */
#define START TEXT0			/* start address */

#include <convex/filehdr.h>
#include <convex/opthdr.h>
#include <convex/scnhdr.h>

extern char environ;

wrtexec (ef)
int ef;
{
    struct filehdr filehdr;
    struct opthdr opthdr;
    struct scnhdr texthdr;
    struct scnhdr datahdr;

    int foffs = 0;
    int ooffs = foffs + sizeof filehdr;
    int toffs = ooffs + sizeof opthdr;
    int doffs = toffs + sizeof texthdr;

    int tsize = DATA0 - TEXT0;
    int dsize = (sbrk (0) - DATA0 + 4095) & -4096;

    bzero (&filehdr, sizeof filehdr);
    bzero (&opthdr, sizeof opthdr);
    bzero (&texthdr, sizeof texthdr);
    bzero (&datahdr, sizeof datahdr);
    
    filehdr.h_magic = SOFF_MAGIC;
    filehdr.h_nscns = 2;
    filehdr.h_scnptr = toffs;
    filehdr.h_opthdr = sizeof opthdr;

    opthdr.o_entry = START;
    opthdr.o_flags = OF_EXEC | OF_STRIPPED;

    texthdr.s_vaddr = TEXT0;
    texthdr.s_size = tsize;
    texthdr.s_scnptr = 0x1000;
    texthdr.s_prot = VM_PG_R | VM_PG_E;
    texthdr.s_flags = S_TEXT;

    datahdr.s_vaddr = DATA0;
    datahdr.s_size = dsize;
    datahdr.s_scnptr = 0x1000 + tsize;
    datahdr.s_prot = VM_PG_R | VM_PG_W;
    datahdr.s_flags = S_DATA;

    write (ef, &filehdr, sizeof filehdr);
    write (ef, &opthdr, sizeof opthdr);
    write (ef, &texthdr, sizeof texthdr);
    write (ef, &datahdr, sizeof datahdr);
    lseek (ef, 0x1000, 0);
    write (ef, TEXT0, tsize + dsize);
    close (ef);

    return dsize;
}
#endif					/* CONVEX */

#ifdef SUN
#include <a.out.h>
wrtexec(ef)
int ef;
{
   struct exec *hdrp, hdr;
   extern environ, etext;
   int tsize, dsize;

   hdrp = (struct exec *)PAGSIZ;
	
   /*
    * This code only handles the ZMAGIC format...
    */
   if (hdrp->a_magic != ZMAGIC)
      syserr("executable is not ZMAGIC format");
   /*
    * Construct the header by copying in the header in core and fixing
    *  up values as necessary.
    */
   hdr = *hdrp;
   tsize = (char *)&etext - (char *)hdrp;
   hdr.a_text = (tsize + PAGSIZ) & ~(PAGSIZ-1);
   dsize = sbrk(0) - (int)&environ;
   hdr.a_data = (dsize + PAGSIZ) & ~(PAGSIZ-1);
   hdr.a_bss = 0;
   hdr.a_syms = 0;
   hdr.a_trsize = 0;
   hdr.a_drsize = 0;

   /*
    * Write the text.
    */
   write(ef, hdrp, tsize);
   lseek(ef, hdr.a_text, 0);

   /*
    * Write the data.
    */
   write(ef, &environ, dsize);
   lseek(ef, hdr.a_data - dsize, 1);

   /*
    * Write the header.
    */
   lseek(ef, 0, 0);
   write(ef, &hdr, sizeof(hdr));

   close(ef);
   return hdr.a_data;
}
#endif					/* SUN */

#ifdef MULTIMAX

#include <time.h>
#include <a.out.h>
#include <sys/file.h>
#include <sgs.h>

#define NUMSECS		2		/* Two sections in the image. */
#define TEXTSTART	0		/* Text starts at address 0. */
#define MODSTART	0x20		/* Depends on crt0.s */
#define MODSIZE		0x10		/* Depends on crt0.s */
#define IMAGEPAGE	4096		/* Page size for images.  Found it */
                                        /* with aoutdump(1) */
#define HDRSIZE		(sizeof(struct filehdr)+sizeof(struct aouthdr)+ \
			 NUMSECS*sizeof(struct scnhdr))
#define PAGEROUND(x)	(((x+IMAGEPAGE-1)/IMAGEPAGE)*IMAGEPAGE)

extern etext;		/* ld(1) puts this at the end of the text segment. */
extern environ;		/* ld(1) puts this at the start of the data segment. */

/*
 * wrtexec() - save image in file.
 */
wrtexec(ExecFile)
     int ExecFile;
{
  int Status;		/* For saving status codes. */

  /* Call internal wrtexec2() routine. */
  Status=wrtexec2(ExecFile);

  /* Close the file. */
  close(ExecFile);
  
  return Status;
}

/*
 * wrtexec2 - Code to write the image file.
 */
static
wrtexec2(ExecFile)
     int ExecFile;
{
  struct filehdr FileHeader;	/* File header record. */
  struct aouthdr SystemHeader;	/* System header record. */
  struct scnhdr SectionHeader;	/* Section header record. */

  struct timeval TV;		/* Time value. */
  struct timezone TZ;		/* Time zone. */

  unsigned long TextStart;	/* Start of text. */
  unsigned long TextSize;	/* Size of text. */
  unsigned long TextFPtr;	/* Location of text in image file. */
  unsigned long DataStart;	/* Start of data. */
  unsigned long DataSize;	/* Size of data. */
  unsigned long DataFPtr;	/* Location of data in image file. */

  /* Figure out a few things we need to know. */
  TextStart = TEXTSTART;
  TextSize = (unsigned long)&etext;
  TextFPtr = PAGEROUND(HDRSIZE);

  DataStart = (unsigned long)&environ;
  DataSize = sbrk(0)-DataStart;
  DataFPtr = TextFPtr+PAGEROUND(TextSize);

  /* Write a file header. */
  FileHeader.f_magic = NS32GMAGIC;		/* NS 32k executable. */
  FileHeader.f_nscns = NUMSECS;			/* Three standard sections. */
  gettimeofday(&TV,&TZ);
  FileHeader.f_timdat = TV.tv_sec;		/* Time stamp. */
  FileHeader.f_symptr = 0;			/* No symbols. */
  FileHeader.f_nsyms = 0;			/* No symbols. */
  FileHeader.f_opthdr = sizeof(struct aouthdr);	/* Size of system header. */
  FileHeader.f_flags = F_RELFLG|F_EXEC|F_LNNO|F_LSYMS; /* Misc. Flags. */
  if(write(ExecFile,&FileHeader,sizeof FileHeader)==-1)
    return -1;

  /* Write a system header. */
  SystemHeader.magic = PAGEMAGIC;		/* Normal executable. */
  SystemHeader.vstamp = 0;			/* Ignore this. */
  SystemHeader.tsize = TextSize;		/* Size of text segment. */
  SystemHeader.dsize = DataSize;		/* Size of data segment. */
  SystemHeader.bsize = 0;			/* No bss */
  SystemHeader.msize = MODSIZE;			/* Magic from aoutdump(1). */
  SystemHeader.mod_start = MODSTART;		/* Magic from aoutdump(1). */
  SystemHeader.entry = 0x2;			/* Magic from aoutdump(1). */
  SystemHeader.text_start = TextStart;		/* Magic from aoutdump(1). */
  SystemHeader.data_start = DataStart;		/* Start of data segment. */
  SystemHeader.entry_mod = 0;			/* Unused. */
  SystemHeader.flags = U_SYS_42|U_AL_4096;	/* UMAX 4.2, 4k align. */
  if(write(ExecFile,&SystemHeader,sizeof SystemHeader)==-1)
    return -1;

  /* Write text section header. */
  strcpy(SectionHeader.s_name,_TEXT);		/* Section name. */
  SectionHeader.s_paddr = TextStart;		/* Physical address. */
  SectionHeader.s_vaddr = TextStart;		/* Virtual address. */
  SectionHeader.s_size = TextSize;		/* Section size. */
  SectionHeader.s_scnptr = TextFPtr;		/* File ptr to section. */
  SectionHeader.s_relptr = 0;			/* No relocation data. */
  SectionHeader.s_lnnoptr = 0;			/* No line numbers. */
  SectionHeader.s_nreloc = 0;			/* No relocation data. */
  SectionHeader.s_nlnno = 0;			/* No line numbers. */
  SectionHeader.s_flags = STYP_TEXT;		/* Text section. */
  SectionHeader.s_symptr = 0;			/* No symbol data. */
  SectionHeader.s_modno = 0;			/* Ignore this. */
  SectionHeader.s_pad = 0;			/* Padding. */
  if(write(ExecFile,&SectionHeader,sizeof SectionHeader)==-1)
    return -1;

  /* Write data section header. */
  strcpy(SectionHeader.s_name,_DATA);		/* Section name. */
  SectionHeader.s_paddr = DataStart;		/* Physical address. */
  SectionHeader.s_vaddr = DataStart;		/* Virtual address. */
  SectionHeader.s_size = DataSize;		/* Section size. */
  SectionHeader.s_scnptr = DataFPtr;		/* File ptr to section. */
  SectionHeader.s_relptr = 0;			/* No relocation data. */
  SectionHeader.s_lnnoptr = 0;			/* No line numbers. */
  SectionHeader.s_nreloc = 0;			/* No relocation data. */
  SectionHeader.s_nlnno = 0;			/* No line numbers. */
  SectionHeader.s_flags = STYP_DATA;		/* Data section. */
  SectionHeader.s_symptr = 0;			/* No symbol data. */
  SectionHeader.s_modno = 0;			/* Ignore this. */
  SectionHeader.s_pad = 0;			/* Padding. */
  if(write(ExecFile,&SectionHeader,sizeof SectionHeader)==-1)
    return -1;

  /* Write the text section. */
  if(lseek(ExecFile,TextFPtr,L_SET)==-1)
    return -1;
  if(write(ExecFile,TextStart,TextSize)==-1)
    return -1;

  /* Write the data section. */
  if(lseek(ExecFile,DataFPtr,L_SET)==-1)
     return -1;
  if(write(ExecFile,DataStart,DataSize)==-1)
    return -1;

  return DataSize;
}

#endif					/* MULTIMAX */

#ifdef GenericBSD
#include <a.out.h>
wrtexec(ef)
int ef;
{
   struct exec hdr;
   extern environ, etext;
   int tsize, dsize;

   /*
    * Construct the header.  The text and data region sizes must be multiples
    *  of 1024.
    */
   hdr.a_magic = ZMAGIC;
   tsize = (int)&etext;
   hdr.a_text = (tsize + 1024) & ~(1024-1);
   dsize = sbrk(0) - (int)&environ;
   hdr.a_data = (dsize + 1024) & ~(1024-1);
   hdr.a_bss = 0;
   hdr.a_syms = 0;
   hdr.a_entry = 0;
   hdr.a_trsize = 0;
   hdr.a_drsize = 0;

   /*
    * Write the header.
    */
   write(ef, &hdr, sizeof(hdr));

   /*
    * Write the text, starting at N_TXTOFF.
    */
   lseek(ef, N_TXTOFF(hdr), 0);
   write(ef, 0, tsize);
   lseek(ef, hdr.a_text - tsize, 1);

   /*
    * Write the data.
    */
   write(ef, &environ, dsize);
   lseek(ef, hdr.a_data - dsize, 1);
   close(ef);
   return hdr.a_data;
}
#endif					/* GenericBSD */

#endif					/* ExecImages */

#if CONVEX

/* replacement pow() that allows negative ** integer */

#undef pow

double pow0 (base, exp)
    double base, exp;
{   if (base >= 0) return pow (base, exp);
    else {
	long n = exp;
	if (n != exp) runerr (-206, 0);
	else if (n & 1) return -pow (-base, exp);
	else return pow (-base, exp);}}
#endif					/* CONVEX */

#endif					/* UNIX */

#if VMS
#include dvidef
#include iodef

typedef struct _descr {
   int length;
   char *ptr;
} descriptor;

typedef struct _pipe {
   long pid;			/* process id of child */
   long status;			/* exit status of child */
   long flags;			/* LIB$SPAWN flags */
   int channel;			/* MBX channel number */
   int efn;			/* Event flag to wait for */
   char mode;			/* the open mode */
   FILE *fptr;			/* file pointer (for fun) */
   unsigned running : 1;	/* 1 if child is running */
} Pipe;

Pipe _pipes[_NFILE];		/* one for every open file */

#define NOWAIT		1
#define NOCLISYM	2
#define NOLOGNAM	4
#define NOKEYPAD	8
#define NOTIFY		16
#define NOCONTROL	32
#define SFLAGS	(NOWAIT|NOKEYPAD|NOCONTROL)

/*
 * popen - open a pipe command
 * Last modified 2-Apr-86/chj
 *
 *	popen("command", mode)
 */

FILE *popen(cmd, mode)
char *cmd;
char *mode;
{
   FILE *pfile;			/* the Pfile */
   Pipe *pd;			/* _pipe database */
   descriptor mbxname;		/* name of mailbox */
   descriptor command;		/* command string descriptor */
   descriptor nl;		/* null device descriptor */
   char mname[65];		/* mailbox name string */
   int chan;			/* mailbox channel number */
   int status;			/* system service status */
   int efn;
   struct {
      short len;
      short code;
      char *address;
      char *retlen;
      int last;
   } itmlst;

   if (!cmd || !mode)
      return (0);
   LIB$GET_EF(&efn);
   if (efn == -1)
      return (0);
   if (_tolower(mode[0]) != 'r' && _tolower(mode[0]) != 'w')
      return (0);
   /* create and open the mailbox */
   status = SYS$CREMBX(0, &chan, 0, 0, 0, 0, 0);
   if (!(status & 1)) {
      LIB$FREE_EF(&efn);
      return (0);
   }
   itmlst.last = mbxname.length = 0;
   itmlst.address = mbxname.ptr = mname;
   itmlst.retlen = &mbxname.length;
   itmlst.code = DVI$_DEVNAM;
   itmlst.len = 64;
   status = SYS$GETDVIW(0, chan, 0, &itmlst, 0, 0, 0, 0);
   if (!(status & 1)) {
      LIB$FREE_EF(&efn);
      return (0);
   }
   mname[mbxname.length] = '\0';
   pfile = fopen(mname, mode);
   if (!pfile) {
      LIB$FREE_EF(&efn);
      SYS$DASSGN(chan);
      return (0);
   }
   /* Save file information now */
   pd = &_pipes[fileno(pfile)];	/* get Pipe pointer */
   pd->mode = _tolower(mode[0]);
   pd->fptr = pfile;
   pd->pid = pd->status = pd->running = 0;
   pd->flags = SFLAGS;
   pd->channel = chan;
   pd->efn = efn;
   /* fork the command */
   nl.length = strlen("_NL:");
   nl.ptr = "_NL:";
   command.length = strlen(cmd);
   command.ptr = cmd;
   status = LIB$SPAWN(&command,
      (pd->mode == 'r') ? 0 : &mbxname,	/* input file */
      (pd->mode == 'r') ? &mbxname : 0,	/* output file */
      &pd->flags, 0, &pd->pid, &pd->status, &pd->efn, 0, 0, 0, 0);
   if (!(status & 1)) {
      LIB$FREE_EF(&efn);
      SYS$DASSGN(chan);
      return (0);
   } else {
      pd->running = 1;
   }
   return (pfile);
}

/*
 * pclose - close a pipe
 * Last modified 2-Apr-86/chj
 *
 */
pclose(pfile)
FILE *pfile;
{
   Pipe *pd;
   int status;
   int fstatus;

   pd = fileno(pfile) ? &_pipes[fileno(pfile)] : 0;
   if (pd == NULL)
      return (-1);
   fflush(pd->fptr);			/* flush buffers */
   fstatus = fclose(pfile);
   if (pd->mode == 'w') {
      status = SYS$QIOW(0, pd->channel, IO$_WRITEOF, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      SYS$WFLOR(pd->efn, 1 << (pd->efn % 32));
   }
   SYS$DASSGN(pd->channel);
   LIB$FREE_EF(&pd->efn);
   pd->running = 0;
   return (fstatus);
}

/*
 * redirect(&argc,argv,nfargs) - redirect standard I/O
 *    int *argc		number of command arguments (from call to main)
 *    char *argv[]	command argument list (from call to main)
 *    int nfargs	number of filename arguments to process
 *
 * argc and argv will be adjusted by redirect.
 *
 * redirect processes a program's command argument list and handles redirection
 * of stdin, and stdout.  Any arguments which redirect I/O are removed from the
 * argument list, and argc is adjusted accordingly.  redirect would typically be
 * called as the first statement in the main program.
 *
 * Files are redirected based on syntax or position of command arguments.
 * Arguments of the following forms always redirect a file:
 *
 *    <file	redirects standard input to read the given file
 *    >file	redirects standard output to write to the given file
 *    >>file	redirects standard output to append to the given file
 *
 * It is often useful to allow alternate input and output files as the
 * first two command arguments without requiring the <file and >file
 * syntax.  If the nfargs argument to redirect is 2 or more then the
 * first two command arguments, if supplied, will be interpreted in this
 * manner:  the first argument replaces stdin and the second stdout.
 * A filename of "-" may be specified to occupy a position without
 * performing any redirection.
 *
 * If nfargs is 1, only the first argument will be considered and will
 * replace standard input if given.  Any arguments processed by setting
 * nfargs > 0 will be removed from the argument list, and again argc will
 * be adjusted.  Positional redirection follows syntax-specified
 * redirection and therefore overrides it.
 *
 */


redirect(argc,argv,nfargs)
int *argc, nfargs;
char *argv[];
{
   int i;

   i = 1;
   while (i < *argc)  {		/* for every command argument... */
      switch (argv[i][0])  {		/* check first character */
         case '<':			/* <file redirects stdin */
            filearg(argc,argv,i,1,stdin,"r");
            break;
         case '>':			/* >file or >>file redirects stdout */
            if (argv[i][1] == '>')
               filearg(argc,argv,i,2,stdout,"a");
            else
               filearg(argc,argv,i,1,stdout,"w");
            break;
         default:			/* not recognized, go on to next arg */
            i++;
      }
   }
   if (nfargs >= 1 && *argc > 1)	/* if positional redirection & 1 arg */
      filearg(argc,argv,1,0,stdin,"r");	/* then redirect stdin */
   if (nfargs >= 2 && *argc > 1)	/* likewise for 2nd arg if wanted */
      filearg(argc,argv,1,0,stdout,"w");/* redirect stdout */
}



/* filearg(&argc,argv,n,i,fp,mode) - redirect and remove file argument
 *    int *argc		number of command arguments (from call to main)
 *    char *argv[]	command argument list (from call to main)
 *    int n		argv entry to use as file name and then delete
 *    int i		first character of file name to use (skip '<' etc.)
 *    FILE *fp		file pointer for file to reopen (typically stdin etc.)
 *    char mode[]	file access mode (see freopen spec)
 */

filearg(argc,argv,n,i,fp,mode)
int *argc, n, i;
char *argv[], mode[];
FILE *fp;
{
   if (strcmp(argv[n]+i,"-"))		/* alter file if arg not "-" */
      fp = freopen(argv[n]+i,mode,fp);
   if (fp == NULL)  {			/* abort on error */
      fprintf(stderr,"%%can't open %s",argv[n]+i);
      exit(ErrorExit);
   }
   for ( ;  n < *argc;  n++)		/* move down following arguments */
      argv[n] = argv[n+1];
   *argc = *argc - 1;			/* decrement argument count */
}

/* Special versions of sbrk() and brk() for use by Icon under VMS.
 * #defines in define.h actually rename these to vms_brk and vms_sbrk.
 *
 * For historical reasons, Icon assumes it can repeatedly call brk/sbrk
 * and always get contiguous chunks.  This was made to work under Unix by
 * overloading the definitions of malloc and friends, the only other callers
 * of sbrk, and making them return Icon-managed memory.

 * Under VMS, sbrk is not the lowest-level system interface.  It gets memory
 * from underlying VMS routines such as SYS$EXPREG.  These routines are also
 * called by others, for example when a file is opened;  so successive sbrk
 * calls may return nonadjacent chunks.  This makes overloading malloc and
 * friends futile.
 *
 * The routines below replace sbrk and brk for Icon (only) under VMS.  They
 * provide the continuously growing memory Icon needs without relying on
 * special privileges or unusually large quotas.  Like the Unix solution and
 * earlier VMS attempts, this is an empirical solution and may need further
 * revision as the system changes.  But we hope not.
 *
 * The Icon interpreter is loaded beginning at address 0 and grows upward as
 * it requests more memory through sbrk.  The C stack grows downward from
 * 0x7FFFFFFF. We're going to draw a line to divide the address space, then
 * force the C and VMS runtime systems to put anything they need above it;
 * then sbrk can grow the program region unimpeded up to the line.
 *
 * The line is drawn MAXMEM bytes beyond the start of the sbrk region.  MAXMEM
 * is an environment variable (logical name to VMS) with a default as given in
 * define.h.  Large values cost CPU and real time expended at process exit; we
 * don't know why.  On an 8600 the cost was very roughly .04 CP sec / megabyte.
 *
 * When first called, sbrk expands the program region by one page to get a
 * starting address.  A limit address is calculated by adding MAXMEM.  A single
 * page created just below the limit address "draws the line" and causes the
 * VMS runtime system to allocate anything it needs above that point.  sbrk
 * creates pages between base and limit as needed.
 *
 * Possible errors and their manifestations:
 *
 *    MAXMEM too large to initialize sbrk:
 *       error in startup code: value of MAXMEM too large
 *
 *    MAXMEM too small to initialize sbrk:
 *       error in startup code: value of MAXMEM too small
 *
 *    MAXMEM too small for subsequent brk/sbrk growth
 *       Run-time error 351:  insufficient MAXMEM limit
 *
 *    MAXMEM okay but insufficient user quota for needed memory:
 *       Run-time error 303:  unable to expand memory region
 *
 *    unexpected ("can't happen") failures of system calls:
 *       these produce their standard VMS error message
 *
 *    unexpected intrusion into the sbrk region by the runtime system:
 *       unknown, but undoubtedly ugly
 */


#define PageSize 512		/* size of a VMS page */
#define MaxP0 0x40000000	/* first address beyond the P0 region */

#include <stsdef.h>

word memsize = MaxMem;		/* set from environment variable MAXMEM */


/*  sbrk(incr) - adjust the break value by incr, rounding up to a page.
 *  returns the new break value, or -1 if unsuccessful.
 */

char *
sbrk(incr)
int incr;
{
   static char *base;		/* base of the sbrk region */
   static char *curr;		/* current break value (end+1) */
   static char *limit;		/* region limit ("the line") */
   char *range[2], *p;		/* scratch for system calls */
   int s;			/* status return from calls */

   /*  initialization code */
   if (!base)  {
      s = sys$expreg(1,range,0,0);	/* expand P0 to get base address */
      if (!(s & STS$M_SUCCESS))
         exit(s);			/* couldn't get one page?! */
      base = curr = range[0];		/* initialize empty sbrk region */
      memsize = (memsize + PageSize - 1) & -PageSize;
					/* round memsize to page boundary */
      limit = base + memsize;		/* calculate sbrk region limit*/
      if (limit > MaxP0)
	 limit = MaxP0;			/* limit to legal values */
      if (limit <= base)
         error("value of MAXMEM too small");  /* can't even start */
      range[0] = range[1] = limit-1;
      s = sys$cretva(range,range,0);	/* get a page there to draw the line */
      if (!(s & STS$M_SUCCESS))
         error("value of MAXMEM too large");  /* can't even start */
   }

   if (incr > 0)  {

      /* grow the region */
      if (curr + incr > limit)		/* check address space available */
         fatalerr(-351,NULL);		/* oops, MAXMEM too small */
      range[0] = curr;
      range[1] = curr + incr - 1;
      s = sys$cretva(range,range,0);	/* ask for the pages */
      if (!(s & STS$M_SUCCESS))
         return (char *) -1;		/* failed, quota exceeded */
      curr = range[1] + 1;		/* set new break value as returned */

   } else if (incr < 0) {

      /* shrink the region (not expected to be used).  does not actually
       * return the memory, but does make it available for reuse.  */
      curr -= -incr & -PageSize;
   }

   /* return the current break value */
   return curr;
}




/*  brk(addr) - set the break address to the given value, rounded up to a page.
 *  returns 0 if successful, -1 if not.
 */

char *
brk(addr)
char *addr;
{
   return ((sbrk(addr-sbrk(0))) == (char *) -1 ? (char *) -1 : 0);
}
#endif					/* VMS */

/*
 * End of operating-system specific code.
 */

char *rjunk;			/* avoid empty module */
